"""Constants."""

RAS_SCHEMA_URL = "https://raw.githubusercontent.com/fema-ffrd/ffrd-conventions/main/hec-ras/json-schema/schema.json"

HMS_SCHEMA_URL = "https://raw.githubusercontent.com/fema-ffrd/ffrd-conventions/main/hec-hms/json-schema/schema.json"
